/*
 * subdirmk - example code
 *  Copyright 2019 Mark Wooding
 * SPDX-License-Identifier: LGPL-2.0-or-later
 * There is NO WARRANTY.
 */

#include <stdio.h>
#include "toylib.h"

int main(void)
  { fputs(greeting(), stdout); return (0); }
