/*
 * subdirmk - example code
 *  Copyright 2019 Mark Wooding
 * SPDX-License-Identifier: LGPL-2.0-or-later
 * There is NO WARRANTY.
 */

#include "toylib.h"

const char *greeting(void)
  { return "Hello, world!\n"; }
