;;; Directory Local Variables
;;; See Info node `(emacs) Directory Variables' for more information.

((c-mode
  (c-basic-offset . 4))
 (python-mode
  (indent-tabs-mode . t)
  (python-indent-offset . 8)
  (python-indent . 8)))
