from setuptools import setup

setup(
    name='base91',
    version='1.0.1',
    packages=['base91'],
    url='https://github.com/SunDwarf/base91-python',
    license='New BSD',
    author='',
    author_email='',
    description='Base91 encoding in Python'
)
